@@ -6,26 +6,34 @@
  *
  * Author:	Miquel van Smoorenburg, miquels@cistron.nl
  *
- * Version:	@(#)last  2.79  13-Jun-2001  miquels@cistron.nl
+ * Version:	@(#)last  2.85  30-Jul-2004  miquels@cistron.nl
  *
  *		This file is part of the sysvinit suite,
- *		Copyright 1991-2001 Miquel van Smoorenburg.
+ *		Copyright (C) 1991-2004 Miquel van Smoorenburg.
  *
- *		This program is free software; you can redistribute it and/or
- *		modify it under the terms of the GNU General Public License
- *		as published by the Free Software Foundation; either version
- *		2 of the License, or (at your option) any later version.
+ *		This program is free software; you can redistribute it and/or modify
+ *		it under the terms of the GNU General Public License as published by
+ *		the Free Software Foundation; either version 2 of the License, or
+ *		(at your option) any later version.
+ *
+ *		This program is distributed in the hope that it will be useful,
+ *		but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *		GNU General Public License for more details.
+ *
+ *		You should have received a copy of the GNU General Public License
+ *		along with this program; if not, write to the Free Software
+ *		Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/fcntl.h>
+#include <fcntl.h>
 #include <time.h>
 #include <stdio.h>
 #include <ctype.h>
 #include <utmp.h>
 #include <errno.h>
-#include <malloc.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
@@ -40,7 +48,11 @@
 #  define SHUTDOWN_TIME 254
 #endif
 
-char *Version = "@(#) last 2.79 11-Sep-2000 miquels";
+#ifndef PATH_MAX
+#define PATH_MAX 2048
+#endif
+
+char *Version = "@(#) last 2.85 31-Apr-2004 miquels";
 
 #define CHOP_DOMAIN	0	/* Define to chop off local domainname. */
 #define NEW_UTMP	1	/* Fancy & fast utmp read code. */
@@ -68,8 +80,12 @@ int maxrecs = 0;	/* Maximum number of records to list. */
 int recsdone = 0;	/* Number of records listed */
 int showhost = 1;	/* Show hostname too? */
 int altlist = 0;	/* Show hostname at the end. */
+int allow_long_username = 0;  /* Show usernames longer than 8 characters */
 int usedns = 0;		/* Use DNS to lookup the hostname. */
 int useip = 0;		/* Print IP address in number format */
+int fulltime = 0;	/* Print full dates and times */
+int name_len = 8;	/* Default print 8 characters of name */
+int domain_len = 16;	/* Default print 16 characters of domain */
 int oldfmt = 0;		/* Use old libc5 format? */
 char **show = NULL;	/* What do they want us to show */
 char *ufile;		/* Filename of this file */
@@ -128,16 +144,16 @@ int uread(FILE *fp, struct utmp *u, int *quit)
 		 *	Initialize and position.
 		 */
 		utsize = oldfmt ? sizeof(uto) : sizeof(struct utmp);
-		fseek(fp, 0L, SEEK_END);
-		fpos = ftell(fp);
+		fseeko(fp, 0, SEEK_END);
+		fpos = ftello(fp);
 		if (fpos == 0)
 			return 0;
 		o = ((fpos - 1) / UCHUNKSIZE) * UCHUNKSIZE;
-		if (fseek(fp, o, SEEK_SET) < 0) {
+		if (fseeko(fp, o, SEEK_SET) < 0) {
 			fprintf(stderr, "%s: seek failed!\n", progname);
 			return 0;
 		}
-		bpos = fpos - o;
+		bpos = (int)(fpos - o);
 		if (fread(buf, bpos, 1, fp) != 1) {
 			fprintf(stderr, "%s: read failed!\n", progname);
 			return 0;
@@ -170,7 +186,7 @@ int uread(FILE *fp, struct utmp *u, int *quit)
 	 *	Copy whatever is left in the buffer.
 	 */
 	memcpy(tmp + (-bpos), buf, utsize + bpos);
-	if (fseek(fp, fpos, SEEK_SET) < 0) {
+	if (fseeko(fp, fpos, SEEK_SET) < 0) {
 		perror("fseek");
 		return 0;
 	}
@@ -207,25 +223,25 @@ int uread(FILE *fp, struct utmp *u, int *quit)
 int uread(FILE *fp, struct utmp *u, int *quit)
 {
 	struct oldutmp uto;
-	int r;
+	off_t r;
 
 	if (u == NULL) {
 		r = oldfmt ? sizeof(struct oldutmp) : sizeof(struct utmp);
-		fseek(fp, -1L * r, SEEK_END);
+		fseek(fp, -1 * r, SEEK_END);
 		return 1;
 	}
 
 	if (!oldfmt) {
 		r = fread(u, sizeof(struct utmp), 1, fp);
 		if (r == 1) {
-			if (fseek(fp, -2L * sizeof(struct utmp), SEEK_CUR) < 0)
+			if (fseeko(fp, -2 * sizeof(struct utmp), SEEK_CUR) < 0)
 				if (quit) *quit = 1;
 		}
 		return r;
 	}
 	r = fread(&uto, sizeof(struct oldutmp), 1, fp);
 	if (r == 1) {
-		if (fseek(fp, -2L * sizeof(struct oldutmp), SEEK_CUR) < 0)
+		if (fseeko(fp, -2 * sizeof(struct oldutmp), SEEK_CUR) < 0)
 			if (quit) *quit = 1;
 		uconv(&uto, u);
 	}
@@ -241,10 +257,11 @@ int uread(FILE *fp, struct utmp *u, int *quit)
 #define BTMP_FILE getbtmp()
 char *getbtmp()
 {
-	static char btmp[128];
+	static char btmp[PATH_MAX + 5];  /* max path + btmp + null terminator */
 	char *p;
 
-	strcpy(btmp, WTMP_FILE);
+        memset(btmp, '\0', PATH_MAX + 5);
+	strncpy(btmp, WTMP_FILE, PATH_MAX);
 	if ((p = strrchr(btmp, '/')) == NULL)
 		p = btmp;
 	else
@@ -300,29 +317,45 @@ char *mybasename(char *s)
 /*
  *	Lookup a host with DNS.
  */
-int dns_lookup(char *result, int size, char *org, unsigned int ip)
+int dns_lookup(char *result, int size, int useip, int32_t *a)
 {
-	struct hostent *h;
+	struct sockaddr_in	sin;
+	struct sockaddr_in6	sin6;
+	struct sockaddr		*sa;
+	int			salen, flags;
+	int			mapped = 0;
+
+	flags = useip ? NI_NUMERICHOST : 0;
 
 	/*
-	 *	Try to catch illegal IP numbers
+	 *	IPv4 or IPv6 ?
+	 *	1. If last 3 4bytes are 0, must be IPv4
+	 *	2. If IPv6 in IPv4, handle as IPv4
+	 *	3. Anything else is IPv6
+	 *
+	 *	Ugly.
 	 */
-	if (ip == 0 || (int)ip == -1 || (ip >> 24) == 0 || (ip & 255) == 0) {
-		if (size > UT_HOSTSIZE) size = UT_HOSTSIZE+1;
-		strncpy(result, org, size - 1);
-		result[size-1] = 0;
-		return 0;
+	if (a[0] == 0 && a[1] == 0 && a[2] == (int32_t)htonl (0xffff))
+		mapped = 1;
+
+	if (mapped || (a[1] == 0 && a[2] == 0 && a[3] == 0)) {
+		/* IPv4 */
+		sin.sin_family = AF_INET;
+		sin.sin_port = 0;
+		sin.sin_addr.s_addr = mapped ? a[3] : a[0];
+		sa = (struct sockaddr *)&sin;
+		salen = sizeof(sin);
+	} else {
+		/* IPv6 */
+		memset(&sin6, 0, sizeof(sin6));
+		sin6.sin6_family = AF_INET6;
+		sin6.sin6_port = 0;
+		memcpy(sin6.sin6_addr.s6_addr, a, 16);
+		sa = (struct sockaddr *)&sin6;
+		salen = sizeof(sin6);
 	}
 
-	if ((h = gethostbyaddr((char *)&ip, 4, AF_INET)) == NULL) {
-		strncpy(result, inet_ntoa(*(struct in_addr *)&ip), size);
-		result[size-1] = 0;
-		return 0;
-	}
-	strncpy(result, h->h_name, size-1);
-	result[size-1] = 0;
-
-	return 0;
+	return getnameinfo(sa, salen, result, size, NULL, 0, flags);
 }
 
 /*
@@ -330,22 +363,22 @@ int dns_lookup(char *result, int size, char *org, unsigned int ip)
  */
 int list(struct utmp *p, time_t t, int what)
 {
-	struct in_addr in;
-	char logintime[32];
-	char logouttime[32];
-	char length[32];
-	char final[128];
-	char utline[UT_LINESIZE+1];
-	char domain[256];
-	time_t secs, tmp;
-	int mins, hours, days;
-	char *s, **walk;
+	time_t		secs, tmp;
+	char		logintime[32];
+	char		logouttime[32];
+	char		length[32];
+	char		final[512];
+	char		utline[UT_LINESIZE+1];
+	char		domain[256];
+	char		*s, **walk;
+	int		mins, hours, days;
+	int		r, len;
 
 	/*
 	 *	uucp and ftp have special-type entries
 	 */
-	strncpy(utline, p->ut_line, UT_LINESIZE);
-	utline[UT_LINESIZE - 1] = 0;
+	utline[0] = 0;
+	strncat(utline, p->ut_line, UT_LINESIZE);
 	if (strncmp(utline, "ftp", 3) == 0 && isdigit(utline[3]))
 		utline[3] = 0;
 	if (strncmp(utline, "uucp", 4) == 0 && isdigit(utline[4]))
@@ -368,10 +401,15 @@ int list(struct utmp *p, time_t t, int what)
 	 *	Calculate times
 	 */
 	tmp = (time_t)p->ut_time;
-	strcpy(logintime, ctime(&tmp));
-	logintime[16] = 0;
-	sprintf(logouttime, "- %s", ctime(&t) + 11);
-	logouttime[7] = 0;
+	strncpy(logintime, ctime(&tmp), sizeof(logintime));
+	logintime[sizeof(logintime)-1] = 0; /* enforce null termination */
+	if (fulltime)
+		sprintf(logouttime, "- %s", ctime(&t));
+	else {
+		logintime[16] = 0;
+		sprintf(logouttime, "- %s", ctime(&t) + 11);
+		logouttime[7] = 0;
+	}
 	secs = t - p->ut_time;
 	mins  = (secs / 60) % 60;
 	hours = (secs / 3600) % 24;
@@ -390,16 +428,23 @@ int list(struct utmp *p, time_t t, int what)
 			break;
 		case R_NOW:
 			length[0] = 0;
-			sprintf(logouttime, "  still");
-			sprintf(length, "logged in");
+			if (fulltime)
+				sprintf(logouttime, "  still logged in");
+			else {
+				sprintf(logouttime, "  still");
+				sprintf(length, "logged in");
+			}
 			break;
 		case R_PHANTOM:
 			length[0] = 0;
-			sprintf(logouttime, "   gone");
-			sprintf(length, "- no logout");
+			if (fulltime)
+				sprintf(logouttime, "  gone - no logout");
+			else {
+				sprintf(logouttime, "   gone");
+				sprintf(length, "- no logout");
+			}
 			break;
 		case R_REBOOT:
-			logouttime[0] = 0;      /* Print machine uptime */
 			break;
 		case R_TIMECHANGE:
 			logouttime[0] = 0;
@@ -412,14 +457,14 @@ int list(struct utmp *p, time_t t, int what)
 	/*
 	 *	Look up host with DNS if needed.
 	 */
-	if (usedns)
-		dns_lookup(domain, sizeof(domain), p->ut_host, p->ut_addr);
-	if (useip) {
-		in.s_addr = p->ut_addr;
-		strcpy(domain, inet_ntoa(in));
-	} else {
-		strncpy(domain, p->ut_host, UT_HOSTSIZE);
-		domain[UT_HOSTSIZE-1] = 0;
+	r = -1;
+	if (usedns || useip)
+		r = dns_lookup(domain, sizeof(domain), useip, p->ut_addr_v6);
+	if (r < 0) {
+		len = UT_HOSTSIZE;
+		if (len >= (int)sizeof(domain)) len = sizeof(domain) - 1;
+		domain[0] = 0;
+		strncat(domain, p->ut_host, len);
 	}
 
 	if (showhost) {
@@ -430,23 +475,75 @@ int list(struct utmp *p, time_t t, int what)
 		if (!usedns && (s = strchr(p->ut_host, '.')) != NULL &&
 		     strcmp(s + 1, domainname) == 0) *s = 0;
 #endif
+#define str(s) # s
+#define xstr(s) str(s)
 		if (!altlist) {
-			snprintf(final, sizeof(final),
-				"%-8.8s %-12.12s %-16.16s "
-				"%-16.16s %-7.7s %-12.12s\n",
-				p->ut_name, utline,
-				domain, logintime, logouttime, length);
+                    if (allow_long_username)  
+                    {
+                       len = snprintf(final, sizeof(final),
+		oldfmt ? "%-" xstr(OLD_NAMESIZE) "." xstr(OLD_NAMESIZE) "s %-12.12s "
+		"%-16.16s %-16.16s %-7.7s %-12.12s\n"
+                : "%-" xstr(UT_NAMESIZE) "." xstr(UT_NAMESIZE) "s %-12.12s "
+		"%-16.16s %-16.16s %-7.7s %-12.12s\n",
+ 				p->ut_name, utline,
+ 				domain, logintime, logouttime, length);
+                    }
+                    else  /* show short username */
+                    {
+			len = snprintf(final, sizeof(final),
+				fulltime ?
+			     "%-8.*s %-12.12s %-16.*s %-24.24s %-26.26s %-12.12s\n" :
+				"%-8.*s %-12.12s %-16.*s %-16.16s %-7.7s %-12.12s\n",
+				name_len, p->ut_name, utline,
+				domain_len, domain, logintime, logouttime, length);
+                    }   /* show short username */
 		} else {
-			snprintf(final, sizeof(final), 
-				"%-8.8s %-12.12s %-16.16s %-7.7s %-12.12s %s\n",
-				p->ut_name, utline,
+                     if (allow_long_username)
+                     {
+                         len = snprintf(final, sizeof(final), 
+		oldfmt ? "%-" xstr(OLD_NAMESIZE) "." xstr(OLD_NAMESIZE) "s %-12.12s "
+		"%-16.16s %-7.7s %-16.16s %s\n"
+	        : "%-" xstr(UT_NAMESIZE) "." xstr(UT_NAMESIZE) "s %-12.12s "
+		"%-16.16s %-7.7s %-16.16s %s\n",
+ 				p->ut_name, utline,
+ 				logintime, logouttime, length, domain);
+                     }
+                     else   /* show short username */
+                     {
+			len = snprintf(final, sizeof(final), 
+				fulltime ?
+				"%-8.*s %-12.12s %-24.24s %-26.26s %-12.12s %s\n" :
+				"%-8.*s %-12.12s %-16.16s %-7.7s %-12.12s %s\n",
+				name_len, p->ut_name, utline,
 				logintime, logouttime, length, domain);
+                     }   /* done showing short username */
 		}
 	} else
-		snprintf(final, sizeof(final),
-			"%-8.8s %-12.12s %-16.16s %-7.7s %-12.12s\n",
-			p->ut_name, utline,
+             if (allow_long_username)
+             {
+                  len = snprintf(final, sizeof(final),
+  		oldfmt ? "%-" xstr(OLD_NAMESIZE) "." xstr(OLD_NAMESIZE) "s %-12.12s "
+		"%-16.16s %-7.7s %-12.12s\n"
+		   : "%-" xstr(UT_NAMESIZE) "." xstr(UT_NAMESIZE) "s %-12.12s "
+		"%-16.16s %-7.7s %-12.12s\n",
+ 			p->ut_name, utline,
+ 			logintime, logouttime, length);
+             }
+             else    /* show short username */
+             {
+		len = snprintf(final, sizeof(final),
+			fulltime ?
+			"%-8.*s %-12.12s %-24.24s %-26.26s %-12.12s\n" :
+			"%-8.*s %-12.12s %-16.16s %-7.7s %-12.12s\n",
+			name_len, p->ut_name, utline,
 			logintime, logouttime, length);
+             }   /* end of showing short username */
+
+#if defined(__GLIBC__)
+#  if (__GLIBC__ == 2) && (__GLIBC_MINOR__ == 0)
+	final[sizeof(final)-1] = '\0';
+#  endif
+#endif
 
 	/*
 	 *	Print out "final" string safely.
@@ -458,6 +555,9 @@ int list(struct utmp *p, time_t t, int what)
 			putchar('*');
 	}
 
+	if (len < 0 || (size_t)len >= sizeof(final))
+		putchar('\n');
+
 	recsdone++;
 	if (maxrecs && recsdone >= maxrecs)
 		return 1;
@@ -472,10 +572,50 @@ int list(struct utmp *p, time_t t, int what)
 void usage(char *s)
 {
 	fprintf(stderr, "Usage: %s [-num | -n num] [-f file] "
-			"[-R] [-x] [-o] [username..] [tty..]\n", s);
+			"[-t YYYYMMDDHHMMSS] "
+			"[-R] [-adioxFw] [username..] [tty..]\n", s);
 	exit(1);
 }
 
+time_t parsetm(char *ts)
+{
+	struct tm	u, origu;
+	time_t		tm;
+
+	memset(&tm, 0, sizeof(tm));
+
+	if (sscanf(ts, "%4d%2d%2d%2d%2d%2d", &u.tm_year,
+	    &u.tm_mon, &u.tm_mday, &u.tm_hour, &u.tm_min,
+	    &u.tm_sec) != 6)
+		return (time_t)-1;
+
+	u.tm_year -= 1900;
+	u.tm_mon -= 1;
+	u.tm_isdst = -1;
+
+	origu = u;
+
+	if ((tm = mktime(&u)) == (time_t)-1)
+		return tm;
+
+	/*
+	 *	Unfortunately mktime() is much more forgiving than
+	 *	it should be.  For example, it'll gladly accept
+	 *	"30" as a valid month number.  This behavior is by
+	 *	design, but we don't like it, so we want to detect
+	 *	it and complain.
+	 */
+	if (u.tm_year != origu.tm_year ||
+	    u.tm_mon != origu.tm_mon ||
+	    u.tm_mday != origu.tm_mday ||
+	    u.tm_hour != origu.tm_hour ||
+	    u.tm_min != origu.tm_min ||
+	    u.tm_sec != origu.tm_sec)
+		return (time_t)-1;
+
+	return tm;
+}
+
 int main(int argc, char **argv)
 {
   FILE *fp;		/* Filepointer of wtmp file */
@@ -499,10 +639,12 @@ int main(int argc, char **argv)
   int extended = 0;	/* Lots of info. */
   char *altufile = NULL;/* Alternate wtmp */
 
+  time_t until = 0;	/* at what time to stop parsing the file */
+
   progname = mybasename(argv[0]);
 
   /* Process the arguments. */
-  while((c = getopt(argc, argv, "f:n:Rxadio0123456789")) != EOF)
+  while((c = getopt(argc, argv, "f:n:RxadFliot:0123456789w")) != EOF)
     switch(c) {
 	case 'R':
 		showhost = 0;
@@ -533,6 +675,25 @@ int main(int argc, char **argv)
 	case 'a':
 		altlist++;
 		break;
+	case 'F':
+		fulltime++;
+		break;
+        case 'l':
+                allow_long_username = 1;
+                break;
+	case 't':
+		if ((until = parsetm(optarg)) == (time_t)-1) {
+			fprintf(stderr, "%s: Invalid time value \"%s\"\n",
+				progname, optarg);
+			usage(progname);
+		}
+		break;
+	case 'w':
+		if (UT_NAMESIZE > name_len)
+			name_len = UT_NAMESIZE;
+		if (UT_HOSTSIZE > domain_len)
+			domain_len = UT_HOSTSIZE;
+		break;
 	case '0': case '1': case '2': case '3': case '4':
 	case '5': case '6': case '7': case '8': case '9':
 		maxrecs = 10*maxrecs + c - '0';
@@ -631,6 +792,9 @@ int main(int argc, char **argv)
 	if (uread(fp, &ut, &quit) != 1)
 		break;
 
+	if (until && until < ut.ut_time)
+		continue;
+
 	if (memcmp(&ut, &oldut, sizeof(struct utmp)) == 0) continue;
 	memcpy(&oldut, &ut, sizeof(struct utmp));
 	lastdate = ut.ut_time;
@@ -648,7 +812,7 @@ int main(int argc, char **argv)
 			ut.ut_type = SHUTDOWN_TIME;
 		else if (strncmp(ut.ut_user, "reboot", 6) == 0)
 			ut.ut_type = BOOT_TIME;
-		else if (strncmp(ut.ut_user, "runlevel", 7) == 0)
+		else if (strncmp(ut.ut_user, "runlevel", 8) == 0)
 			ut.ut_type = RUN_LVL;
 	}
 #if 1 /*def COMPAT*/
@@ -682,8 +846,8 @@ int main(int argc, char **argv)
 	switch (ut.ut_type) {
 		case SHUTDOWN_TIME:
 			if (extended) {
-				strcpy(ut.ut_line, "system down");
-				quit = list(&ut, lastdown, R_NORMAL);
+				strncpy(ut.ut_line, "system down", OLD_LINESIZE - 1);
+				quit = list(&ut, lastboot, R_NORMAL);
 			}
 			lastdown = lastrch = ut.ut_time;
 			down = 1;
@@ -691,15 +855,16 @@ int main(int argc, char **argv)
 		case OLD_TIME:
 		case NEW_TIME:
 			if (extended) {
-				strcpy(ut.ut_line,
+				strncpy(ut.ut_line,
 				ut.ut_type == NEW_TIME ? "new time" :
-					"old time");
+					"old time", OLD_LINESIZE - 1);
 				quit = list(&ut, lastdown, R_TIMECHANGE);
 			}
 			break;
 		case BOOT_TIME:
-			strcpy(ut.ut_line, "system boot");
+			strncpy(ut.ut_line, "system boot", OLD_LINESIZE - 1);
 			quit = list(&ut, lastdown, R_REBOOT);
+			lastboot = ut.ut_time;
 			down = 1;
 			break;
 		case RUN_LVL:
